/*
                                          [Source file] 12/25/2003

'WineWorksShortcutResolver' is a class to parse Windows shortcut (.lnk) files.

This class is pure C++ and should compile with any compilant compiler.
This class was written by me for WineWorks, but you are free to use it in your program provided your software is GPLed too.

An alternative license may be made available if requested.

Incase you notice BUGs, send a detailed report (and breakpoint if applicable) to <subhobrotosinha at yahoo.com>

Copyright (C) 2004 Subhobroto Sinha <subhobrotosinha at yahoo.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Parts which are (C) WINE, are applicable under the current WINE License (LGPL).
(However, devoted WINE coders will find that the functions are heavily modified ;) Hence, this code will NOT work if simply linked to the original WINE source!)

NOTE: If GetPath() returns "Desktop", it means that it's a Desktop link.
I really don't know what to do with such links on a GNU System... (Ordinarily on a WINDOWS box, you just look up the process name under CLSID registry entry)
*/

#include "WineWorksShortcutResolver.h"
#include <wchar.h>

#define SHOWMETHOSEMESSAGES 0
///////////////////////////////////////////////////////////////////////////////////////
/*
The following two functions are required to copy wchar_t (which is 4 bytes on GNU) to WCHAR (which is 2 bytes on Windows)
Though Windows typedefs WCHAR as a wchar_t, but there a wchar_t is an (unsigned short), whereas on GNU it's (unsigned long)
So WCHAR serialized data from Windows will get messed if we use native GNU data types.

If any side effects of my assumptions are noticed, alert me immediately please.
*/
inline void Convert_wchar_t_ToWCHAR(const wchar_t* wszIn,WCHAR* wszOut)
{
  if(wszIn) while(*wszOut++=*wszIn++);
  else *wszOut=0;
}
inline void ConvertWCHARTo_wchar_t(const WCHAR* wszIn,wchar_t* wszOut)
{
  if(wszIn) while(*wszOut++=*wszIn++);
  else *wszOut=0;
}
///////////////////////////////////////////////////////////////////////////////////////

void Error(const char* error,BOOL bTerminate=false);
inline unsigned int UnicodeLen(const WCHAR* wszIn)
{
  /*
  Internal function to get length of a Unicode string, including the terminating NULL.
  */
  wchar_t wszOut[10*1024]={0};
  ConvertWCHARTo_wchar_t(wszIn,wszOut);
  return (wszIn?(wcslen(wszOut)+1):0);
  //return (wszIn?(wcslen(wszIn)+1):0);
}

WineWorksShortcutResolver::WineWorksShortcutResolver()
{
	memset(&wthdtMe,0,sizeof(wthdtMe));
}

WineWorksShortcutResolver::WineWorksShortcutResolver(const char* lpszFileName)
{
	oifstrStr.open(lpszFileName/*,ios::binary|ios::in|ios::nocreate*/);
	memset(&wthdtMe,0,sizeof(wthdtMe));
	LoadShortcut();
}

WineWorksShortcutResolver::~WineWorksShortcutResolver()
{
    oifstrStr.close();
    if(wthdtMe.sIcoPath) delete[] wthdtMe.sIcoPath;
    if(wthdtMe.sArgs) delete[] wthdtMe.sArgs;
    if(wthdtMe.sWorkDir) delete[] wthdtMe.sWorkDir;
    if(wthdtMe.sDescription) delete[] wthdtMe.sDescription;
    if(wthdtMe.sPath) delete[] wthdtMe.sPath;
    if(wthdtMe.pPidl) delete[] wthdtMe.pPidl;

}

LPPIDLDATA WineWorksShortcutResolver::_dbg_ILGetDataPointer(LPCITEMIDLIST pidl)
{
	if(pidl&&pidl->mkid.cb!=0x00) return (LPPIDLDATA)&(pidl->mkid.abID);
	return NULL;
}
///////////////////////////////////////////////////////////////////////////
LPITEMIDLIST WineWorksShortcutResolver::_dbg_ILGetNext(LPCITEMIDLIST pidl)
{
	return (pidl&&pidl->mkid.cb ? (LPITEMIDLIST)(((LPBYTE)pidl)+pidl->mkid.cb):NULL);
}
///////////////////////////////////////////////////////////////////////////
BOOL WineWorksShortcutResolver::_dbg_ILIsDesktop(LPCITEMIDLIST pidl)
{
	return (!pidl||(pidl&&pidl->mkid.cb==0x00));
}
///////////////////////////////////////////////////////////////////////////
LPSTR WineWorksShortcutResolver::_dbg_ILGetTextPointer(LPCITEMIDLIST pidl)
{
	LPPIDLDATA pdata =_dbg_ILGetDataPointer(pidl);

	if (pdata)
	{
	  switch (pdata->type)
	  {
	    case PT_MYCOMP:
	    case PT_SPECIAL:
			return NULL;

	    case PT_DRIVE:
	    case PT_DRIVE1:
	    case PT_DRIVE2:
	    case PT_DRIVE3:
			return (LPSTR)(pdata->u.drive.szDriveName-3);

	    case PT_FOLDER:
	    case PT_FOLDER1:
	    case PT_VALUE:
	    case PT_IESPECIAL1:
	    case PT_IESPECIAL2:
			return (LPSTR)(pdata->u.file.szNames-6);

	    case PT_WORKGRP:
	    case PT_COMP:
	    case PT_NETWORK:
	    case PT_NETPROVIDER:
	    case PT_SHARE:
			return (LPSTR)(pdata->u.network.szNames-6);
	  }
	}
	return NULL;
}
///////////////////////////////////////////////////////////////////////////
DWORD WineWorksShortcutResolver::_dbg_ILSimpleGetText (LPCITEMIDLIST pidl, LPSTR szOut, UINT uOutSize)
{
	LPSTR szSrc=_dbg_ILGetTextPointer(pidl);
	
	if(!pidl) return 0;
	if(szOut) *szOut=0;

	if(_dbg_ILIsDesktop(pidl))
	{
	 /* desktop */
	  if(szOut) strncpy(szOut,"Desktop",uOutSize);
	  return strlen("Desktop");
	}
	else if(szSrc)
	{
	  /* filesystem */
	  if(szOut) strncpy(szOut,szSrc,uOutSize);
	  return strlen(szSrc);
	}
	return 0;
}
///////////////////////////////////////////////////////////////////////////
unsigned int WineWorksShortcutResolver::UnicodeToANSI(const WCHAR* wszInUnicode,unsigned int uiLen,char* szOutANSI)
{
  /*
  uiLen is length of wszInUnicode
  szOutANSI is buffer uiANSILen long (to accept the converted string..)
  returns the number of number of characters copied from wszInUnicode, and -1 in case of an error
  */
  wchar_t wszT[uiLen];
  ConvertWCHARTo_wchar_t(wszInUnicode,wszT);
  return wszInUnicode?wcstombs(szOutANSI,wszT,uiLen):(*szOutANSI=0);//*szOutANSI=0 should make szOutANSI empty and then return 0 implying no characters were copied.
  //return wszInUnicode?wcstombs(szOutANSI,wszInUnicode,uiLen):(*szOutANSI=0);//*szOutANSI=0 should make szOutANSI empty and then return 0 implying no characters were copied.
}
///////////////////////////////////////////////////////////////////////////
unsigned int WineWorksShortcutResolver::ANSIToUnicode(const char* szInANSI,unsigned int uiANSILen,WCHAR* wszOutUnicode)
{
  /*
  uiANSILen is length of szInANSI
  wszOutUnicode is buffer atleast 2*uiANSILen long (to accept the converted string..)
  returns the number of number of characters copied from szInANSI, and -1 in case of an error
  */
  wchar_t wszT[uiANSILen];
  mbstowcs(wszT,szInANSI,(uiANSILen+1));
  Convert_wchar_t_ToWCHAR(wszT,wszOutUnicode);
  return szInANSI?wcslen(wszT):(*wszOutUnicode=0);//*wszOutUnicode=0 should make wszOutUnicode empty and then return 0 implying no characters were copied.
  //return szInANSI?mbstowcs(wszOutUnicode,szInANSI,(uiANSILen+1)):(*wszOutUnicode=0);//*wszOutUnicode=0 should make wszOutUnicode empty and then return 0 implying no characters were copied.
}

///////////////////////////////////////////////////////////////////////////
HRESULT WineWorksShortcutResolver::Stream_LoadLocation()
{
    DWORD size=0;

    oifstrStr.read((char*)&size,sizeof(size));
    
    LOCATION_INFO *loc=new LOCATION_INFO[size];
    if(!loc) return -1;

    oifstrStr.read((char *)&loc->dwHeaderSize,size-sizeof(size));
    loc->dwTotalSize = size;
	
    /* FIXME: do something useful with it
    Just to move the file pointer... ;)*/
    delete[] loc;
    loc=NULL;
    return S_OK;
}
///////////////////////////////////////////////////////////////////////////
HRESULT WineWorksShortcutResolver::Stream_LoadString(LPWSTR& pstr)
{
  ULONG len=0;
  pstr=NULL;

  oifstrStr.read((char*)&len,sizeof(USHORT));
  len *= sizeof(WCHAR);//This gives just the length of the string, EXCLUDING the NULL terminator.	
  pstr=new WCHAR[len+sizeof(WCHAR)];//The '+sizeof(WCHAR)' adds the size for the terminating NULL
  if(!pstr) return -1;
  oifstrStr.read((char *)pstr,len);
  /////////////////////////////////////////////////////////////////////////////////////
  len/=2;
  char* szTemp=new char[len+1];
  UnicodeToANSI(pstr,len,szTemp);
  szTemp[len]='\0';
  //Convert szTemp to Unicode.szTemp contains the correct ANSI string
  ANSIToUnicode(szTemp,len,pstr);
  delete[] szTemp;
  /////////////////////////////////////////////////////////////////////////////////////
  return S_OK;
}
///////////////////////////////////////////////////////////////////////////
BOOL WineWorksShortcutResolver::LoadShortcut()
{
	char szTemp[256]={0};
	LINK_HEADER *pLnkHrd=new LINK_HEADER[sizeof(LINK_HEADER)];
	if(!oifstrStr.is_open()) Error("File error",true);
	oifstrStr.read((char*)pLnkHrd,sizeof(LINK_HEADER));
	if((pLnkHrd->MagicStr!=0x0000004CL)) Error("Not a shortcut",true);
	WORD wLen=0;
	oifstrStr.read((char*)&wLen,2);
#if SHOWMETHOSEMESSAGES
	sprintf(szTemp,"PIDL length is %d",wLen);
	Error(szTemp);
#endif
  	if(wLen!=0)
	{
		wthdtMe.pPidl=new ITEMIDLIST[wLen];
		ITEMIDLIST *pPidl=new ITEMIDLIST[wLen];
		oifstrStr.read((char*)pPidl,wLen);
		LPITEMIDLIST pidltemp=_dbg_ILGetNext(pPidl);
		char szPasha[128][MAX_PATH]={0};
		int i=0,j=1;
		do
		{
			_dbg_ILSimpleGetText(pidltemp,szPasha[i++],MAX_PATH);
			if(!strcmp(szPasha[0],"Desktop")) break;
			pidltemp = _dbg_ILGetNext(pidltemp);
		}while(pidltemp->mkid.cb);
		delete[] wthdtMe.pPidl;
		wthdtMe.pPidl=NULL;
		//////////////////////////////
		strcpy(szTemp,szPasha[0]);
		for(;j<i;j++)
		{
			strcat(szTemp,szPasha[j]);
			if(j!=(i-1))strcat(szTemp,"\\");
		}
		strcat(szTemp,szPasha[j]);
		wthdtMe.sPath=new WCHAR[strlen(szTemp)+1];
		//szTemp has the correct ANSI path
		ANSIToUnicode(szTemp,(strlen(szTemp)+1),wthdtMe.sPath);
#if SHOWMETHOSEMESSAGES
		Error(szTemp);
#endif
		//////////////////////////////
    HRESULT r;
    if(pLnkHrd->dwFlags & SCF_NORMAL) r = Stream_LoadLocation();
		if(pLnkHrd->dwFlags & SCF_DESCRIPTION) r = Stream_LoadString(wthdtMe.sDescription);
		if(pLnkHrd->dwFlags & SCF_RELATIVE ) r = Stream_LoadString(wthdtMe.sPathRel);
		if(pLnkHrd->dwFlags & SCF_WORKDIR )	r = Stream_LoadString(wthdtMe.sWorkDir);
		if(pLnkHrd->dwFlags & SCF_ARGS ) r = Stream_LoadString(wthdtMe.sArgs);
		if(pLnkHrd->dwFlags & SCF_CUSTOMICON ) r = Stream_LoadString(wthdtMe.sIcoPath);
	}
	delete[] pLnkHrd;
	pLnkHrd=NULL;
	return 0;
}

void Error(const char* error,BOOL bTerminate)
{
  fputs(error,stdout);
  fputs("\n",stdout);
  if(bTerminate) exit(1);
}
/*
The followings are the worker functions which you are likely to use in your program(s).
Each expects a pointer to a string preallocated from your program which will accept the requested data
Returns the number of characters left uncopied from source to destination. -1 indicates an error


The usage of this class is very simple:

  (1) Define an object of this class and pass the filename of the shortcut to the constructor.
  (2) Call any/all of the following functions with a preallocated string as argument to get the required data
  
NOTE: The passed buffer should be large enough to hold the converted string !
*/

DWORD  WineWorksShortcutResolver::GetPath(char* lpszPath)
{
	if(!lpszPath) return false;
	int iLen=UnicodeLen(wthdtMe.sPath);
	int iT=(iLen-UnicodeToANSI(wthdtMe.sPath,iLen,lpszPath));
	return(iT>iLen?-1:iT);
}

DWORD  WineWorksShortcutResolver::GetArgs(char* lpszArgs)
{
	if(!lpszArgs) return false;
	int iLen=UnicodeLen(wthdtMe.sArgs);
	int iT=(iLen-UnicodeToANSI(wthdtMe.sArgs,iLen,lpszArgs));
	return(iT>iLen?-1:iT);
}

DWORD  WineWorksShortcutResolver::GetWorkDir(char* lpszWorkDir)
{
	if(!lpszWorkDir) return false;
	int iLen=UnicodeLen(wthdtMe.sWorkDir);
	int iT=(iLen-UnicodeToANSI(wthdtMe.sWorkDir,iLen,lpszWorkDir));
	return(iT>iLen?-1:iT);
}

DWORD  WineWorksShortcutResolver::GetDescription(char* lpszDescription)
{
	if(!lpszDescription) return false;
	int iLen=UnicodeLen(wthdtMe.sDescription);
	int iT=(iLen-UnicodeToANSI(wthdtMe.sDescription,iLen,lpszDescription));
	return(iT>iLen?-1:iT);
}

DWORD  WineWorksShortcutResolver::GetPathRel(char* lpszPathRel)
{
	if(!lpszPathRel) return false;
	int iLen=UnicodeLen(wthdtMe.sPathRel);
	int iT=(iLen-UnicodeToANSI(wthdtMe.sPathRel,iLen,lpszPathRel));
	return(iT>iLen?-1:iT);
}
